@@ -13,7 +13,7 @@
 ##                                                               #
 ##  Written by Stepan Tsirkin, University of Zurich.             #
 ##  e-mail: stepan.tsirkin@physik.uzh.ch                         #
-##################################################################{0:8.4f}   {1:8.4f}  ".format(
+##################################################################
 
 
 import copy
@@ -22,14 +22,88 @@ import functools
 import numpy as np
 import numpy.linalg as la
 
-from .__aux import str2bool, BOHR
-from .__readfiles import AbinitHeader, Hartree_eV
-from .__readfiles import WAVECARFILE
+from .aux import str2bool, BOHR
+from .readfiles import AbinitHeader, Hartree_eV
+from .readfiles import WAVECARFILE
 from .kpoint import Kpoint
 from .spacegroup import SpaceGroup
 
 
 class BandStructure:
+    """
+    Parses files and organizes info about the whole band structure in 
+    attributes. Contains methods to calculate and write traces (and irreps), 
+    for the separation of the band structure in terms of a symmetry operation 
+    and for the calculation of the Zak phase and wannier charge centers.
+
+    Parameters
+    ----------
+    fWAV : str, default=None
+        Name of file containing wave-functions in VASP (WAVECAR format).
+    fWFK : str, default=None
+        Name of file containing wave-functions in ABINIT (WFK format).
+    prefix : str, default=None
+        Prefix used for Quantum Espresso calculations or seedname of Wannier90 
+        files.
+    fPOS : str, default=None
+        Name of file containing the crystal structure in VASP (POSCAR format).
+    Ecut : float, default=None
+        Plane-wave cutoff in eV to consider in the expansion of wave-functions.
+    IBstart : int, default=None
+        First band to be considered.
+    IBend : int, default=None
+        Last band to be considered.
+    kplist : array, default=None
+        List of indices of k-points to be considered.
+    spinor : bool, default=None
+        `True` if wave functions are spinors, `False` if they are scalars. 
+        Mandatory for VASP.
+    code : str, default='vasp'
+        DFT code used. Set to 'vasp', 'abinit', 'espresso' or 'wannier90'.
+    EF : float, default=None
+        Fermi-energy.
+    onlysym : bool, default=False
+        Exit after printing info about space-group.
+    spin_channel : str, default=None
+        Selection of the spin-channel. 'up' for spin-up, 'dw' for spin-down. 
+        Only applied in the interface to Quantum Espresso.
+    refUC : array, default=None
+        3x3 array describing the transformation of vectors defining the 
+        unit cell to the standard setting.
+    shiftUC : array
+        Translation taking the origin of the unit cell used in the DFT 
+        calculation to that of the standard setting.
+
+    Attributes
+    ----------
+    spacegroup : class
+        Instance of `SpaceGroup`.
+    spinor : bool
+        `True` if wave functions are spinors, `False` if they are scalars. It 
+        will be read from DFT files.
+    efermi : float
+        Fermi-energy. If user set a number as `EF` in CLI, it will be used. If 
+        `EF` was set to `auto`, it will try to parse it and set to 0.0 if it 
+        could not.
+    Ecut0 : float
+        Plane-wave cutoff (in eV) used in DFT calulations. Always read from 
+        DFT files. Insignificant if `code`='wannier90'.
+    Ecut : float
+        Plane-wave cutoff (in eV) to consider in the expansion of wave-functions.
+        Will be set equal to `Ecut0` if input parameter `Ecut` was not set or 
+        the value of this is negative or larger than `Ecut0`.
+    Lattice : array, shape=(3,3) 
+        Each row contains cartesian coordinates of a basis vector forming the 
+        unit-cell in real space.
+    RecLattice : array, shape=(3,3)
+        Each row contains the cartesian coordinates of a basis vector forming 
+        the unit-cell in reciprocal space.
+    kpoints : list
+        Each element is an instance of `class Kpoint` corresponding to a 
+        k-point specified in input parameter `kpoints`. If this input was not 
+        set, all k-points found in DFT files will be considered.
+    """
+
     def __init__(
         self,
         fWAV=None,
@@ -42,9 +116,11 @@ class BandStructure:
         kplist=None,
         spinor=None,
         code="vasp",
-        EF=None,
+        EF='0.0',
         onlysym=False,
-        spin_channel=None
+        spin_channel=None,
+        refUC = None,
+        shiftUC = None
     ):
         code = code.lower()
 
@@ -54,19 +130,19 @@ class BandStructure:
 
         if code == "vasp":
             self.__init_vasp(
-                fWAV, fPOS, Ecut, IBstart, IBend, kplist, spinor, EF=EF, onlysym=onlysym
+                fWAV, fPOS, Ecut, IBstart, IBend, kplist, spinor, EF=EF, onlysym=onlysym, refUC=refUC, shiftUC=shiftUC
             )
         elif code == "abinit":
             self.__init_abinit(
-                fWFK, Ecut, IBstart, IBend, kplist, EF=EF, onlysym=onlysym
+                fWFK, Ecut, IBstart, IBend, kplist, EF=EF, onlysym=onlysym, refUC=refUC, shiftUC=shiftUC
             )
         elif code == "espresso":
             self.__init_espresso(
-                prefix, Ecut, IBstart, IBend, kplist, EF=EF, onlysym=onlysym, spin_channel=spin_channel
+                prefix, Ecut, IBstart, IBend, kplist, EF=EF, onlysym=onlysym, spin_channel=spin_channel, refUC=refUC, shiftUC=shiftUC
             )
         elif code == "wannier90":
             self.__init_wannier(
-                prefix, Ecut, IBstart, IBend, kplist, EF=EF, onlysym=onlysym
+                prefix, Ecut, IBstart, IBend, kplist, EF=EF, onlysym=onlysym, refUC=refUC, shiftUC=shiftUC
             )
         else:
             raise RuntimeError("Unknown/unsupported code :{}".format(code))
@@ -80,19 +156,63 @@ class BandStructure:
         IBend=None,
         kplist=None,
         spinor=None,
-        EF=None,
+        EF='0.0',
         onlysym=False,
+        refUC=None,
+        shiftUC=None,
     ):
+        """
+        Initialization for vasp. Read data and save it in attributes.
+
+        Parameters
+        ----------
+        fWAV : str, default=None
+            Filename for wavefunction in VASP WAVECAR format.
+        fPOS : str, default=None
+            Filename for wavefunction in VASP POSCAR format.
+        Ecut : float, default=None
+            Plane-wave cutoff in eV to consider in the expansion of wave-functions.
+        IBstart : int, default=None
+            First band to be considered.
+        IBend : int, default=None
+            Last band to be considered.
+        kplist : , default=None
+            List of indices of k-points to be considered.
+        spinor : bool, default=None
+            `True` if wave functions are spinors, `False` if they are scalars.
+        EF : str, default='0.0'
+            Fermi-energy.
+        onlysym : bool, default=False
+            Exit after printing info about space-group.
+        refUC : array, default=None
+            3x3 array describing the transformation of vectors defining the 
+            unit cell to the standard setting.
+        shiftUC : array, default=None
+            Translation taking the origin of the unit cell used in the DFT 
+            calculation to that of the standard setting.
+        """
         if spinor is None:
             raise RuntimeError(
                 "spinor should be specified in the command line for VASP bandstructure"
             )
-        self.spacegroup = SpaceGroup(inPOSCAR=fPOS, spinor=spinor)
+        self.spacegroup = SpaceGroup(inPOSCAR=fPOS, spinor=spinor, refUC=refUC, shiftUC=shiftUC)
         self.spinor = spinor
         if onlysym:
             return
-        self.efermi = 0.0 if EF is None else EF
-        print("Efermi = ", self.efermi, EF)
+
+        if EF.lower() == "auto":
+            self.efermi = 0.0
+            msg = " (Fermi-energy not found in WAVECAR)"
+        else:
+            try:
+                self.efermi = float(EF)
+                msg = ""
+            except:
+                raise RuntimeError(
+                        ("Invalid value for keyword EF. It must be "
+                         "a number or 'auto'")
+                        )
+        print("Efermi = {:.4f} eV".format(self.efermi) + msg)
         WCF = WAVECARFILE(fWAV)
         # RECLENGTH=3 # the length of a record in WAVECAR. It is defined in the
         # first record, so let it be 3 fo far"
@@ -158,7 +278,7 @@ class BandStructure:
                 Ecut,
                 Ecut0,
                 self.RecLattice,
-                SG=self.spacegroup,
+                symmetries_SG=self.spacegroup.symmetries,
                 spinor=self.spinor,
                 WCF=WCF,
             )
@@ -172,19 +292,58 @@ class BandStructure:
         IBstart=None,
         IBend=None,
         kplist=None,
-        EF=None,
+        EF='0.0',
         onlysym=False,
+        refUC=None,
+        shiftUC=None,
     ):
+        """
+        Initialization for abinit. Read data and store it in attributes.
+
+        Parameters
+        ----------
+        WFKname : str
+            Filename for wavefunction in ABINIT WFK format.
+        Ecut : float, default=None
+            Plane-wave cutoff in eV to consider in the expansion of wave-functions.
+        IBstart : int, default=None
+            First band to be considered.
+        IBend : int, default=None
+            Last band to be considered.
+        kplist : , default=None
+            List of indices of k-points to be considered.
+        EF : str, default='0.0'
+            Fermi-energy.
+        onlysym : bool, default=False
+            Exit after printing info about space-group.
+        refUC : array, default=None
+            3x3 array describing the transformation of vectors defining the 
+            unit cell to the standard setting.
+        shiftUC : array, default=None
+            Translation taking the origin of the unit cell used in the DFT 
+            calculation to that of the standard setting.
+        """
 
         header = AbinitHeader(WFKname)
         usepaw = header.usepaw
         self.spinor = header.spinor
         self.spacegroup = SpaceGroup(
-            cell=(header.rprimd, header.xred, header.typat), spinor=self.spinor
+            cell=(header.rprimd, header.xred, header.typat), spinor=self.spinor, refUC=refUC, shiftUC=shiftUC
         )
         if onlysym:
             return
-        self.efermi = header.efermi * Hartree_eV if EF is None else EF
+        if EF.lower() == "auto":
+            self.efermi = header.efermi * Hartree_eV
+        else:
+            try:
+                self.efermi = float(EF)
+            except:
+                raise RuntimeError(
+                        ("Invalid value for keyword EF. It must be "
+                         "a number or 'auto'")
+                        )
+        print("Efermi: {:.4f} eV".format(self.efermi))
+
         #        self.spacegroup.show()
 
         #        global fWFK
@@ -241,7 +400,7 @@ class BandStructure:
                 Ecut,
                 Ecut0,
                 self.RecLattice,
-                SG=self.spacegroup,
+                symmetries_SG=self.spacegroup.symmetries,
                 spinor=self.spinor,
                 code="abinit",
                 kpt=header.kpt[ik],
@@ -260,16 +419,52 @@ class BandStructure:
         IBstart=None,
         IBend=None,
         kplist=None,
-        EF=None,
+        EF='0.0',
         onlysym=False,
+        refUC=None,
+        shiftUC=None,
     ):
+        """
+        Initialization for wannier90. Read data and store it in attibutes.
 
+        Parameters
+        ----------
+        prefix : str, default=None
+            Prefix used for Quantum Espresso calculations or seedname of Wannier90 
+            files.
+        Ecut : float, default=None
+            Plane-wave cutoff in eV to consider in the expansion of wave-functions.
+        IBstart : int, default=None
+            First band to be considered.
+        IBend : int, default=None
+            Last band to be considered.
+        kplist : , default=None
+            List of indices of k-points to be considered.
+        EF : str, default='0.0'
+            Fermi-energy.
+        onlysym : bool, default=False
+            Exit after printing info about space-group.
+        refUC : array, default=None
+            3x3 array describing the transformation of vectors defining the 
+            unit cell to the standard setting.
+        shiftUC : array, default=None
+            Translation taking the origin of the unit cell used in the DFT 
+            calculation to that of the standard setting.
+        """
         if Ecut is None:
             raise RuntimeError("Ecut mandatory for Wannier90")
 
         fwin = [l.strip().lower() for l in open(prefix + ".win").readlines()]
 
         def split(l):
+            """
+            Determine symbol used for assignment and split accordingly.
+
+            Parameters
+            ---------
+            l : str
+                Part of a line read from .win file.
+            """
             if "=" in l:
                 return l.split("=")
             elif ":" in l:
@@ -286,6 +481,36 @@ class BandStructure:
         # print(fwin) #com
 
         def get_param(key, tp, default=None, join=False):
+            """
+            Return value of a parameter in .win file.
+
+            Parameters
+            ----------
+            key : str
+                Wannier90 input parameter.
+            tp : function
+                Function to apply to the value of the parameter, before 
+                returning it.
+            default
+                Default value to return in case parameter `key` is not found.
+            join : bool, default=False
+                If the value of parameter `key` contains more than one element, 
+                they will be concatenated with a blank space if `join` is set 
+                to `True`. Used when the parameter is `mpgrid`.
+
+            Returns
+            -------
+            Type(`tp`)
+                Return the value of the parameter, after applying function 
+                passed es keyword `tp`.
+
+            Raises
+            ------
+            RuntimeError
+                The parameter is not found in .win file, it is found more than 
+                once or its value is formed by many elements but it is not
+                `mpgrid`.
+            """
             i = np.where(ind == key)[0]
             if len(i) == 0:
                 if default is None:
@@ -317,20 +542,23 @@ class BandStructure:
         NBin = get_param("num_bands", int)
         #        print ("nbands=",NBin)
         self.spinor = str2bool(get_param("spinors", str))
-        if EF is None:
+
+        if EF.lower() == "auto":
             try:
                 self.efermi = (
-                    get_param("fermi_energy", float, 0.0) if EF is None else EF
-                )
-            except Exception as err:
-                print(
-                    "WARNING : fermi_energy not found.Setting as zero : `{}`".format(
-                        err
+                    get_param("fermi_energy", float, 0.0)
                     )
-                )
-                self.efermi = 0
+            except:
+                print("WARNING : fermi-energy not found. Setting it as zero")
         else:
-            self.efermi = EF
+            try:
+                self.efermi = float(EF)
+            except:
+                raise RuntimeError(
+                        ("Invalid value for keyword EF. It must be "
+                         "a number or 'auto'")
+                        )
+        print("Efermi = {:.4f} eV".format(self.efermi))
 
         NK = np.prod(np.array(get_param("mp_grid", str).split(), dtype=int))
 
@@ -341,13 +569,26 @@ class BandStructure:
             kplist = np.arange(NK) + 1
         else:
             # kplist-=1 #files start from 1 in W90
-            kplist = np.array([k for k in kplist if k >= 0 and k < NK])
+            kplist = np.array([k for k in kplist if k > 0 and k <= NK])
         found_atoms = False
         # todo : use an iterator to avoid double looping over lines between
         # "begin" and "end"
         iterwin = iter(fwin)
 
         def check_end(name):
+            """
+            Check if block in .win file is closed.
+
+            Parameters
+            ----------
+            name : str
+                Name of the block in .win file.
+            
+            Raises
+            ------
+            RuntimeError
+                Block is not closed.
+            """
             s = next(iterwin)
             if " ".join(s) != "end " + name:
                 raise RuntimeError(
@@ -370,7 +611,7 @@ class BandStructure:
                         L = [next(iterwin) for i in range(3)]
                     else:
                         units = "ang"
-                        L = l1 + [next(iterwin)[0] for i in range(2)]
+                        L = [l1] + [next(iterwin) for i in range(2)]
                     self.Lattice = np.array(L, dtype=float)
                     if units == "bohr":
                         self.Lattice *= BOHR
@@ -415,7 +656,7 @@ class BandStructure:
                         xred.append(l1[1:4])
                     typatdic = {n: i + 1 for i, n in enumerate(set(nameat))}
                     typat = [typatdic[n] for n in nameat]
-                    xred = np.array(xred)
+                    xred = np.array(xred, dtype=float)
                     if l[1][6:10] == "cart":
                         xred = xred.dot(np.linalg.inv(self.Lattice))
 
@@ -433,7 +674,7 @@ class BandStructure:
         )
 
         self.spacegroup = SpaceGroup(
-            cell=(self.Lattice, xred, typat), spinor=self.spinor
+            cell=(self.Lattice, xred, typat), spinor=self.spinor, refUC=refUC, shiftUC=shiftUC
         )
         if onlysym:
             return
@@ -475,7 +716,7 @@ class BandStructure:
                 Ecut,
                 None,
                 self.RecLattice,
-                SG=self.spacegroup,
+                symmetries_SG=self.spacegroup.symmetries,
                 spinor=self.spinor,
                 code="wannier",
                 eigenval=eigenval[ik - 1],
@@ -491,10 +732,41 @@ class BandStructure:
         IBstart=None,
         IBend=None,
         kplist=None,
-        EF=None,
+        EF='0.0',
         onlysym=False,
-        spin_channel=None
+        spin_channel=None,
+        refUC=None,
+        shiftUC=None,
     ):
+        """
+        Initialization for Quantum Espresso. Read data and store in attributes.
+
+        Parameters
+        ----------
+        prefix : str, default=None
+            Prefix used for Quantum Espresso calculations or seedname of Wannier90 
+            files.
+        Ecut : float, default=None
+            Plane-wave cutoff in eV to consider in the expansion of wave-functions.
+        IBstart : int, default=None
+            First band to be considered.
+        IBend : int, default=None
+            Last band to be considered.
+        kplist : , default=None
+            List of indices of k-points to be considered.
+        EF : str, default='0.0'
+            Fermi-energy.
+        onlysym : bool, default=False
+            Exit after printing info about space-group.
+        spin_channel : str, default=None
+            Selection of the spin-channel. 'up' for spin-up, 'dw' for spin-down.
+        refUC : array, default=None
+            3x3 array describing the transformation of vectors defining the 
+            unit cell to the standard setting.
+        shiftUC : array, default=None
+            Translation taking the origin of the unit cell used in the DFT 
+            calculation to that of the standard setting.
+        """
         import xml.etree.ElementTree as ET
 
         mytree = ET.parse(prefix + ".save/data-file-schema.xml")
@@ -528,7 +800,7 @@ class BandStructure:
         xred = (np.array(xcart, dtype=float) * BOHR).dot(np.linalg.inv(self.Lattice))
         #        print ("xred=",xred)
         self.spacegroup = SpaceGroup(
-            cell=(self.Lattice, xred, typat), spinor=self.spinor
+            cell=(self.Lattice, xred, typat), spinor=self.spinor, refUC=refUC, shiftUC=shiftUC
         )
         if onlysym:
             return
@@ -584,11 +856,26 @@ class BandStructure:
             * np.pi
             / np.linalg.det(self.Lattice)
         )
-        try:
-            self.efermi = float(bandstr.find("fermi_energy").text) * Hartree_eV
-        except Exception as err:
-            print("WARNING : fermi_energy not found.Setting as zero : `{}`".format(err))
-            self.efermi = 0
+        
+        if EF.lower() == "auto":
+            try:
+                self.efermi = (
+                               float(bandstr.find("fermi_energy").text) 
+                               * Hartree_eV
+                               )
+            except:
+                print("WARNING : fermi-energy not found. Setting it as zero")
+                self.efermi = 0.0
+        else:
+            try:
+                self.efermi = float(EF)
+            except:
+                raise RuntimeError(
+                        ("Invalid value for keyword EF. It must be "
+                         "a number or 'auto'")
+                        )
+        print("Efermi: {:.4f} eV".format(self.efermi))
+
         kpall = bandstr.findall("ks_energies")
         NK = len(kpall)
         if kplist is None:
@@ -610,7 +897,7 @@ class BandStructure:
                 Ecut,
                 Ecut0,
                 self.RecLattice,
-                SG=self.spacegroup,
+                symmetries_SG=self.spacegroup.symmetries,
                 spinor=self.spinor,
                 code="espresso",
                 kptxml=kpall[ik],
@@ -629,24 +916,56 @@ class BandStructure:
     #        exit()
 
     def getNK():
+        """Getter for `self.kpoints`."""
         return len(self.kpoints)
 
     NK = property(getNK)
 
     def write_characters(
         self,
-        degen_thresh=0,
-        refUC=None,
-        shiftUC=np.zeros(3),
+        degen_thresh=1e-8,
         kpnames=None,
         symmetries=None,
         preline="",
         plotFile=None,
     ):
+        """
+        Calculate irreps, number of band-inversion (if little-group contains 
+        inversion), smallest direct gap and indirect gap and print all of them.
+
+        Parameters
+        ----------
+        degen_thresh : float, default=1e-8
+            Threshold energy used to decide whether wave-functions are
+            degenerate in energy.
+        refUC : array, default=None
+            3x3 array describing the transformation of vectors defining the 
+            unit cell to the standard setting.
+        shiftUC : array, default=np.zeros(3)
+            Translation taking the origin of the unit cell used in the DFT 
+            calculation to that of the standard setting.
+        kpnames : list, default=None
+            Labels of maximal k-points at which irreps of bands must be computed. 
+            If it is not specified, only traces will be printed, not irreps.
+        symmetries : list, default=None
+            Index of symmetry operations whose description will be printed. 
+        preline : str, default=''
+            Characters to write before labels of irreps in file `irreps.dat`.
+        plotFile : str, default=None
+            Name of file in which energy-levels and corresponding irreps will be 
+            written to later place irreps in a band structure plot.
+
+        Returns
+        -------
+        json_data : `json` object
+            Object with output structured in `json` format.
+        """
         #        if refUC is not None:
         #        self.spacegroup.show(refUC=refUC,shiftUC=shiftUC)
         #        self.spacegroup.show2(refUC=refUC)
         kpline = self.KPOINTSline()
+        json_data = {}
+        json_data[ "kpoints_line"] = kpline
         try:
             pFile = open(plotFile, "w")
         except BaseException:
@@ -655,12 +974,11 @@ class BandStructure:
         GAP = np.Inf
         Low = -np.Inf
         Up = np.inf
-        if kpnames is not None and refUC is not None:
+        json_data["k-points" ] = []
+        if kpnames is not None:
             for kpname, KP in zip(kpnames, self.kpoints):
-                irreps = self.spacegroup.get_irreps_from_table(
-                    refUC, shiftUC, kpname, KP.K
-                )
-                ninv, low, up = KP.write_characters(
+                irreps = self.spacegroup.get_irreps_from_table(kpname, KP.K)
+                ninv, low, up , kdata = KP.write_characters(
                     degen_thresh,
                     irreptable=irreps,
                     symmetries=symmetries,
@@ -668,14 +986,20 @@ class BandStructure:
                     efermi=self.efermi,
                     plotFile=pFile,
                     kpl=kpline,
+                    symmetries_tables=self.spacegroup.symmetries_tables,
+                    refUC=self.spacegroup.refUC,
+                    shiftUC=self.spacegroup.shiftUC
                 )
+                kdata["kpname"] = kpname
+                json_data["k-points" ].append(kdata)
+
                 NBANDINV += ninv
                 GAP = min(GAP, up - low)
                 Up = min(Up, up)
                 Low = max(Low, low)
         else:
             for KP, kpl in zip(self.kpoints, kpline):
-                ninv, low, up = KP.write_characters(
+                ninv, low, up , kdata = KP.write_characters(
                     degen_thresh,
                     symmetries=symmetries,
                     preline=preline,
@@ -683,6 +1007,8 @@ class BandStructure:
                     plotFile=pFile,
                     kpl=kpl,
                 )
+                kdata["kp in line"] = kpl
+                json_data["k-points" ].append(kdata)
                 NBANDINV += ninv
                 GAP = min(GAP, up - low)
                 Up = min(Up, up)
@@ -695,26 +1021,40 @@ class BandStructure:
                 "  Z4= ",
                 int(NBANDINV / 2) % 4,
             )
+            json_data["number of inversions-odd Kramers pairs"]  = int(NBANDINV / 2)
+            json_data["Z4"] = int(NBANDINV / 2) % 4,
         else:
             print("number of inversions-odd states : ", NBANDINV)
+            json_data["number of inversions-odd states"]  = NBANDINV
 
         #        print ("Total number of inversion-odd Kramers pairs IN THE LISTED KPOINTS: ",NBANDINV,"  Z4= ",NBANDINV%4)
         print("Minimal direct gap:", GAP, " eV")
         print("indirect  gap:", Up - Low, " eV")
+        json_data["indirect gap (eV)"] =  Up-Low
+        json_data["Minimal direct gap (eV)"] =  GAP
+       
+        return json_data
 
     def getNbands(self):
-        # returns the number of bands (if equal over all k-points) of
-        # RuntimeError otherwise
+        """
+        Return number of bands (if equal over all k-points), raise RuntimeError 
+        otherwise.
+
+        Returns
+        -------
+        int
+            Number of bands in every k-point.
+        """
         nbarray = [k.Nband for k in self.kpoints]
         if len(set(nbarray)) > 1:
             raise RuntimeError(
-                "the numbers of bands differs over k-points:{0} \n cannot write tracce.txt \n".format(
+                "the numbers of bands differs over k-points:{0} \n cannot write trace.txt \n".format(
                     nbarray
                 )
             )
         if len(nbarray) == 0:
             raise RuntimeError(
-                "do we have any k-points??? NB={0} \n cannot write tracce.txt \n".format(
+                "do we have any k-points??? NB={0} \n cannot write trace.txt \n".format(
                     nbarray
                 )
             )
@@ -722,12 +1062,25 @@ class BandStructure:
 
     def write_trace(
         self,
-        degen_thresh=0,
-        refUC=None,
-        shiftUC=np.zeros(3),
+        degen_thresh=1e-8,
         kpnames=None,
         symmetries=None,
     ):
+        """
+        Generate `trace.txt` file to upload to the program `CheckTopologicalMat` 
+        in `BCS <https://www.cryst.ehu.es/cgi-bin/cryst/programs/topological.pl>`_ .
+
+        Parameters
+        ----------
+        degen_thresh : float, default=1e-8
+            Threshold energy used to decide whether wave-functions are
+            degenerate in energy.
+        kpnames : list, default=None
+            Labels of maximal k-points at which irreps of bands must be computed. 
+            If it is not specified, only traces will be printed, not irreps.
+        symmetries : list, default=None
+            Index of symmetry operations whose description will be printed. 
+        """
         f = open("trace.txt", "w")
         f.write(
             (
@@ -736,7 +1089,9 @@ class BandStructure:
             ).format(self.getNbands(), 1 if self.spinor else 0)
         )
 
-        f.write(self.spacegroup.write_trace(refUC=refUC, shiftUC=shiftUC))
+        f.write(
+                self.spacegroup.write_trace()
+                )
         # Number of maximal k-vectors in the space group. In the next files
         # introduce the components of the maximal k-vectors))
         f.write("  {0}  \n".format(len(self.kpoints)))
@@ -744,7 +1099,7 @@ class BandStructure:
             f.write(
                 "   ".join(
                     "{0:10.6f}".format(x)
-                    for x in (refUC.dot(KP.K) if refUC is not None else KP.K)
+                    for x in self.spacegroup.refUC.dot(KP.K)
                 )
                 + "\n"
             )
@@ -754,15 +1109,35 @@ class BandStructure:
             )
 
     def Separate(self, isymop, degen_thresh=1e-5, groupKramers=True):
+        """
+        Separate band structure according to the eigenvalues of a symmetry 
+        operation.
+        
+        Parameters
+        ----------
+        isymop : int
+            Index of symmetry used for the separation.
+        degen_thresh : float, default=1e-5
+            Energy threshold used to determine degeneracy of energy-levels.
+        groupKramers : bool, default=True
+            If `True`, states will be coupled by Kramers' pairs.
+
+        Returns
+        -------
+        subspaces : dict
+            Each key is an eigenvalue of the symmetry operation and the
+            corresponding value is an instance of `class` `BandStructure` for 
+            the subspace of that eigenvalue.
+        """
         if isymop == 1:
             return {1: self}
         symop = self.spacegroup.symmetries[isymop - 1]
-        print("Separating by symmetry operation # ", isymop)
+        #print("Separating by symmetry operation # ", isymop)
         symop.show()
         kpseparated = [
             kp.Separate(symop, degen_thresh=degen_thresh, groupKramers=groupKramers)
             for kp in self.kpoints
-        ]
+        ] # each element is a dict with separated bandstructure of a k-point
         allvalues = np.array(sum((list(kps.keys()) for kps in kpseparated), []))
         #        print (allvalues)
         #        for kps in kpseparated :
@@ -781,7 +1156,7 @@ class BandStructure:
             if len(borders) > 2:
                 allvalues = set(
                     [allvalues[b1:b2].mean() for b1, b2 in zip(borders, borders[1:])]
-                )
+                ) # unrepeated Re parts of all eigenvalues
                 subspaces = {}
                 for v in allvalues:
                     other = copy.copy(self)
@@ -791,7 +1166,7 @@ class BandStructure:
                         vk0 = vk[np.argmin(np.abs(v - vk))]
                         if abs(vk0 - v) < 0.05:
                             other.kpoints.append(K[vk0])
-                        subspaces[v] = other
+                    subspaces[v] = other # unnecessary indent ?
                 return subspaces
             else:
                 return dict({allvalues.mean(): self})
@@ -825,6 +1200,25 @@ class BandStructure:
                 return dict({allvalues.mean(): self})
 
     def zakphase(self):
+        """
+        Calculate Zak phases along a path for a set of states.
+
+        Returns
+        -------
+        z : array
+            `z[i]` contains the total  (trace) zak phase (divided by 
+            :math:`2\pi`) for the subspace of the first i-bands.
+        array
+            The :math:`i^{th}` element is the gap between :math:`i^{th}` and
+            :math:`(i+1)^{th}` bands in the considered set of bands. 
+        array
+            The :math:`i^{th}` element is the mean energy between :math:`i^{th}` 
+            and :math:`(i+1)^{th}` bands in the considered set of bands. 
+        array
+            Each line contains the local gaps between pairs of bands in a 
+            k-point of the path. The :math:`i^{th}` column is the local gap 
+            between :math:`i^{th}` and :math:`(i+1)^{th}` bands.
+        """
         overlaps = [
             x.overlap(y)
             for x, y in zip(self.kpoints, self.kpoints[1:] + [self.kpoints[0]])
@@ -835,6 +1229,8 @@ class BandStructure:
         print("   sum  ", np.sum(np.angle(O[0, 0]) for O in overlaps) / np.pi)
         #        overlaps.append(self.kpoints[-1].overlap(self.kpoints[0],g=np.array( (self.kpoints[-1].K-self.kpoints[0].K).round(),dtype=int )  )  )
         nmax = np.min([o.shape for o in overlaps])
+        # calculate zak phase in incresing dimension of the subspace (1 band,
+        # 2 bands, 3 bands,...)
         z = np.angle(
             [[la.det(O[:n, :n]) for n in range(1, nmax + 1)] for O in overlaps]
         ).sum(axis=0) % (2 * np.pi)
@@ -856,6 +1252,15 @@ class BandStructure:
         return z, emin - emax, (emin + emax) / 2, locgap
 
     def wcc(self):
+        """
+        Calculate Wilson loops.
+
+        Returns
+        -------
+        array
+            Eigenvalues of the Wilson loop operator, divided by :math:`2\pi`.
+
+        """
         overlaps = [
             x.overlap(y)
             for x, y in zip(self.kpoints, self.kpoints[1:] + [self.kpoints[0]])
@@ -868,6 +1273,16 @@ class BandStructure:
         return np.sort((np.angle(np.linalg.eig(wilson)) / (2 * np.pi)) % 1)
 
     def write_bands(self, locs=None):
+        """
+        Generate lines for a band structure plot, with cummulative length of the
+        k-path as values for the x-axis and energy-levels for the y-axis.
+
+        Returns
+        -------
+        str
+            Lines to write into a file that will be parsed to plot the band 
+            structure.
+        """
         #        print (locs)
         kpline = self.KPOINTSline()
         nbmax = max(k.Nband for k in self.kpoints)
@@ -899,26 +1314,38 @@ class BandStructure:
 
     def write_trace_all(
         self,
-        degen_thresh=0,
-        refUC=None,
-        shiftUC=np.zeros(3),
+        degen_thresh=1e-8,
         symmetries=None,
         fname="trace_all.dat",
     ):
+        """
+        Write in a file the description of symmetry operations, energy-levels 
+        and irreps calculated in all k-points.
+
+        Parameters
+        ----------
+        degen_thresh : float, default=1e-8
+            Threshold energy used to decide whether wave-functions are
+            degenerate in energy.
+        symmetries : list, default=None
+            Index of symmetry operations whose traces will be printed. 
+        fname : str, default=trace_all.dat
+            Name of output file.
+        """
         f = open(fname, "w")
         kpline = self.KPOINTSline()
+
         f.write(
             (
                 "# {0}  # Number of bands below the Fermi level\n"
                 + "# {1}  # Spin-orbit coupling. No: 0, Yes: 1\n"  #
             ).format(self.getNbands(), 1 if self.spinor else 0)
         )
+        # add lines describing symmetry operations
         f.write(
             "\n".join(
                 ("#" + l)
-                for l in self.spacegroup.write_trace(
-                    refUC=refUC, shiftUC=shiftUC
-                ).split("\n")
+                for l in self.spacegroup.write_trace().split("\n")
             )
             + "\n\n"
         )
@@ -929,7 +1356,27 @@ class BandStructure:
                 )
             )
 
+
     def KPOINTSline(self, kpred=None, breakTHRESH=0.1):
+        """
+        Calculate cumulative length along a path in reciprocal space.
+
+        Parameters
+        ----------
+        kpred : list, default=None
+            Each element contains the direct coordinates of a k-point in the
+            attribute `kpoints`.
+        breakTHRESH : float, default=0.1
+            If the distance between two neighboring k-points in the path is 
+            larger than `breakTHRESH`, it is taken to be 0.
+
+        Returns
+        -------
+        array
+            Each element is the cumulative distance along the path up to a 
+            k-point. The first element is 0, so that the number of elements
+            matches the number of k-points in the path.
+        """
         if kpred is None:
             kpred = [k.K for k in self.kpoints]
         KPcart = np.array(kpred).dot(self.RecLattice)
